/*	$LAAS: t3dcomp.c,v 1.2 2005/08/31 21:56:52 sara Exp $ */

/*
 * Copyright (c) 2000-2004 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "t3d/t3d.h"
#include "t3d_internal.h"


/* Local data -------------------------------------------------------- */

static T3D *	t3dCompMATRIX(T3D *r, T3D *a, T3D *b);
static T3D *	t3dCompUMATRIX(T3D *r, T3D *a, T3D *b);


/* --- t3dComp ------------------------------------------------------- */

/** \brief Compose two T3D.
 *
 * Compose T3D a with T3D b, that is compute a T3D that is the
 * combination of first applying a and then b. The result is stored in a
 * and a is returned.
 *
 * If a represents the transformation from a frame A to a frame X and b
 * represents the transformation from a frame X to a frame B, the result of
 * this fonction is a transformation from the frame A to the frame B.
 * Or for short : t3Comp("A_2_X", "X_2_B") = "A_2_B".
 *
 * Note that if T3D has an internal type of T3D_MATRIX, the result of
 * this function is the product of matrix b by matrix a.
 *
 * \param[out] a	First transformation.
 * \param[in] b		Second transformation, applied after a.
 *
 * \return a updated.
 */

T3D *
t3dComp(T3D *a, T3D *b)
{
  T3D r;

  r.flags = a->flags;
  if (!t3dCompIn(&r, a, b)) return NULL;
  return t3dCopyIn(a, &r);
}


/* --- t3dCompIn ----------------------------------------------------- */

/** \brief Compose two T3D.
 *
 * Compose T3D a with T3D b, store the result in c and return c.
 * See t3dComp().
 *
 * \param[out] c	Resulting transformation.
 * \param[in] a		First transformation.
 * \param[in] b		Second transformation, applied after a.
 *
 * \return c.
 */

T3D *
t3dCompIn(T3D *c, T3D *a, T3D *b)
{
  if (!t3dOk(a)) { t3dWarn(a); return a; }
  if (!t3dOk(b)) { t3dWarn(b); return b; }
  t3dSetOk(c);

  /* Handle all type compositions through the matrix representation type
   * since the direct function would require to compute the matrices
   * anyway. */

  if (!t3dUncertain(a->type) && !t3dUncertain(b->type))
    return t3dCompMATRIX(c, a, b);
  else
    return t3dCompUMATRIX(c, a, b);
}


/* --- r3dComp ------------------------------------------------------- */

/** \brief Compose two R3D.
 *
 * Compose R3D a with T3D b, store the result in c and return c.
 * See t3dComp().
 *
 * \param[out] a	First transformation.
 * \param[in] b		Second transformation, applied after a.
 *
 * \return a, updated.
 */

R3D *
r3dComp(R3D *a, R3D *b)
{
  return (R3D *)t3dComp((T3D *)a, (T3D *)b);
}


/* --- r3dCompIn ----------------------------------------------------- */

/** \brief Compose two R3D.
 *
 * Compose R3D a with T3D b, store the result in c and return c.
 * See t3dComp().
 *
 * \param[out] c	Resulting transformation.
 * \param[in] a		First transformation.
 * \param[in] b		Second transformation, applied after a.
 *
 * \return c.
 */

R3D *
r3dCompIn(R3D *c, R3D *a, R3D *b)
{
  return (R3D *)t3dCompIn((T3D *)c, (T3D *)a, (T3D *)b);
}


/* --- t3dCompMATRIX ------------------------------------------------- */

/* Compose two T3D by multiplying homogeneous matrices. Note that to
 * respect the semantics of t3dComp(a, b) where a is the first
 * transformation and b the second, b's matrix must be pre-multiplied. */

static T3D *
t3dCompMATRIX(T3D *r, T3D *a, T3D *b)
{
  if (a->type != T3D_MATRIX) {
    if (a->flags & T3D_ALLOW_CONVERSION)
      t3dConvertTo(T3D_MATRIX, a);
    else
      return t3dSetError(r, T3D_ENOTIMPLEMENTED);
  }

  if (b->type != T3D_MATRIX) {
    if (b->flags & T3D_ALLOW_CONVERSION)
      t3dConvertTo(T3D_MATRIX, b);
    else
      return t3dSetError(r, T3D_ENOTIMPLEMENTED);
  }

  /* multiply matrices */
  t3d_dm_mult(b->matrix.matrix, a->matrix.matrix, r->matrix.matrix, 4, 4, 4, 4);
  r->type = T3D_MATRIX;
  t3dSetOk(r);
  return r;
}


/*
 * --- compMatrices -----------------------------------------------------
 *
 * Compose two T3D_MATRIX, and variances on Euler parameters.
 */

static T3D *
t3dCompUMATRIX(T3D *r, T3D *a, T3D *b)
{
   double *c;				/* resulting matrix */
   double eulera[3], eulerb[3];		/* Euler angles of a and b */
   double rota[9], rotb[9];		/* rotation matrices of a and b */

   double d, d2;
   double Jee1[36], Jee2[36];
   double JeaR[27], JRea[27];
   double Jeaea1[9], Jeaea2[9];
   double JRa[9], JRb[9], JRg[9];
   double Jta[3], Jtb[3], Jtg[3];
   double Jtea2[9];
   double tmp1[21], tmp2[21];

   if (a->type != T3D_U_MATRIX) {
     if (a->flags & T3D_ALLOW_CONVERSION)
       t3dConvertTo(T3D_U_MATRIX, a);
     else
       return t3dSetError(r, T3D_ENOTIMPLEMENTED);
   }

   if (b->type != T3D_U_MATRIX) {
     if (b->flags & T3D_ALLOW_CONVERSION)
       t3dConvertTo(T3D_U_MATRIX, b);
     else
       return t3dSetError(r, T3D_ENOTIMPLEMENTED);
   }

   /* compute Euler angles of a and b */
   t3dEulerAnglesOfMatrix(a->umatrix.matrix, eulera);
   t3dEulerAnglesOfMatrix(b->umatrix.matrix, eulerb);

   /* extract rotation matrices from a and b */
   t3dRotationOfMatrix(a->umatrix.matrix, rota);
   t3dRotationOfMatrix(b->umatrix.matrix, rotb);

   /* compose matrices */
   c = r->umatrix.matrix;
   t3d_dm_mult(b->umatrix.matrix, a->umatrix.matrix, c, 4, 4, 4, 4);


   /*
    * Now variances. If ea are the Euler angles, we just say that
    *     Dea = JeaR JRea1 Dea1 + JeaR JRea2 Dea2
    */


   /* first, the jacobian JeaR */
   d = sqrt(d2 = (c[0] * c[0] + c[4] * c[4]));

   if (d2 > 1e-5) {
      double s = c[10] * c[10] + c[9] * c[9];
      t3d_dm_fill(
	 JeaR, 3, 9,
	 -c[4]/d2,    0.0, 0.0, c[0]/d2,     0.0, 0.0, 0.0, 0.0,     0.0,
	 c[8]*c[0]/d, 0.0, 0.0, c[8]*c[4]/d, 0.0, 0.0, -d,  0.0,     0.0,
	 0.0,         0.0, 0.0, 0.0,         0.0, 0.0, 0.0, c[10]/s, -c[9]/s
	 );

   } else
      t3d_dm_fill(
	 JeaR, 3, 9,
	 0.0,         -c[5], 0.0, 0.0,         c[1], 0.0, 0.0, 0.0, 0.0,
	 c[8]*c[0]/d, 0.0,   0.0, c[8]*c[4]/d, 0.0,  0.0, -d,  0.0, 0.0,
	 0.0,         0.0,   0.0, 0.0,         0.0,  0.0, 0.0, 0.0, 0.0
	 );

   /* then, the two from the rotation matrices  R = R2 R1 */

   t3dDerivativeOfEuler(eulera, JRa, JRb, JRg);
   t3d_dm_mult(rotb, JRa, tmp1, 3, 3, 3, 3); memcpy(JRa, tmp1, sizeof(JRa));
   t3d_dm_mult(rotb, JRb, tmp1, 3, 3, 3, 3); memcpy(JRb, tmp1, sizeof(JRb));
   t3d_dm_mult(rotb, JRg, tmp1, 3, 3, 3, 3); memcpy(JRg, tmp1, sizeof(JRg));

   t3d_dm_blocfill(JRea, 9, 3, 9, 1, JRa, JRb, JRg);
   t3d_dm_mult(JeaR, JRea, Jeaea1, 3, 9, 9, 3);

   /* translation t = R2.t1 + t2, so dt/dea1 = 0 and dt/dea2 = dR2/dea2 * t1 */
   tmp1[0] = a->umatrix.matrix[3];
   tmp1[1] = a->umatrix.matrix[7];
   tmp1[2] = a->umatrix.matrix[11];

   t3dDerivativeOfEuler(eulerb, JRa, JRb, JRg);
   t3d_dm_mult(JRa, tmp1, Jta, 3, 3, 3, 1);
   t3d_dm_mult(JRb, tmp1, Jtb, 3, 3, 3, 1);
   t3d_dm_mult(JRg, tmp1, Jtg, 3, 3, 3, 1);
   t3d_dm_blocfill(Jtea2, 3, 3, 3, 1, Jta, Jtb, Jtg);

   t3d_dm_mult(JRa, rota, tmp1, 3, 3, 3, 3); memcpy(JRa, tmp1, sizeof(JRa));
   t3d_dm_mult(JRb, rota, tmp1, 3, 3, 3, 3); memcpy(JRb, tmp1, sizeof(JRb));
   t3d_dm_mult(JRg, rota, tmp1, 3, 3, 3, 3); memcpy(JRg, tmp1, sizeof(JRg));

   t3d_dm_blocfill(JRea, 9, 3, 9, 1, JRa, JRb, JRg);
   t3d_dm_mult(JeaR, JRea, Jeaea2, 3, 9, 9, 3);

   /* the two jacobian matrices */
   t3d_dm_blocfill(Jee1, 6, 6, 3, 3, Jeaea1, NULL, NULL, rotb);
   t3d_dm_blocfill(Jee2, 6, 6, 3, 3, Jeaea2, NULL, Jtea2, t3dmid3);

   /* variance: if C = B o A , V(C) = J^C_A V(A) J^C_A + J^C_B V(B) J^C_B */
   t3d_dmfsf_jmult_6x6(Jee1, a->umatrix.v, tmp1);
   t3d_dmfsf_jmult_6x6(Jee2, b->umatrix.v, tmp2);
   t3d_dms_plus_6x6(tmp1, tmp2, r->umatrix.v);

   r->type = T3D_U_MATRIX;
   t3dSetOk(r);
   return r;
}
