/*	$LAAS: t3dconvert.c,v 1.2 2005/08/31 21:56:52 sara Exp $ */

/*
 * Copyright (c) 2000-2004 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "t3d/t3d.h"
#include "t3d_internal.h"


/* --- local data ---------------------------------------------------- */

static T3D *	t3dConvertToU_(T3D *t3d);
static T3D *	t3dConvertToC_(T3D *t3d);

static T3D *	t3dConvertMATRIXtoBRYAN(T3D *t3d);
static T3D *	t3dConvertMATRIXtoVECTOR(T3D *t3d);
static T3D *	t3dConvertU_MATRIXtoU_BRYAN(T3D *t3d);
static T3D *	t3dConvertU_MATRIXtoU_VECTOR(T3D *t3d);

static T3D *	t3dConvertBRYANtoMATRIX(T3D *t3d);
static T3D *	t3dConvertU_BRYANtoU_MATRIX(T3D *t3d);

static T3D *	t3dConvertVECTORtoMATRIX(T3D *t3d);
static T3D *	t3dConvertU_VECTORtoU_MATRIX(T3D *t3d);


/* --- t3dConvertTo -------------------------------------------------- */

/** \brief Convert t3d to the "typeTo" representation.
 *
 * \param[in] typeTo	Type to convert to.
 * \param[in] t3d	T3D object to convert.
 *
 * \return t3d, converted.
 */

T3D *
t3dConvertTo(T3DType typeTo, T3D *t3d)
{
   if (!t3dOk(t3d)) { t3dWarn(t3d); return t3d; }
   if (t3d->type == typeTo) return t3d;

   /* first convert from certain/uncertain type */
   if (t3dUncertain(typeTo) != t3dUncertain(t3d->type)) {
     if (t3dUncertain(typeTo))
       t3dConvertToU_(t3d);
     else
       t3dConvertToC_(t3d);
   }
   if (t3d->type == typeTo) return t3d;

   switch(t3dMixedType(t3d->type, typeTo)) {
     /* From MATRIX */
     case t3dMixedType(T3D_MATRIX, T3D_BRYAN):
       return t3dConvertMATRIXtoBRYAN(t3d);

     case t3dMixedType(T3D_MATRIX, T3D_VECTOR):
       return t3dConvertMATRIXtoVECTOR(t3d);

     case t3dMixedType(T3D_U_MATRIX, T3D_U_BRYAN):
       return t3dConvertU_MATRIXtoU_BRYAN(t3d);

     case t3dMixedType(T3D_U_MATRIX, T3D_U_VECTOR):
       return t3dConvertU_MATRIXtoU_VECTOR(t3d);

     /* From BRYAN */
     case t3dMixedType(T3D_BRYAN, T3D_MATRIX):
       return t3dConvertBRYANtoMATRIX(t3d);

     case t3dMixedType(T3D_BRYAN, T3D_VECTOR):
       t3dConvertBRYANtoMATRIX(t3d);
       return t3dConvertMATRIXtoVECTOR(t3d);

     case t3dMixedType(T3D_U_BRYAN, T3D_U_MATRIX):
       return t3dConvertU_BRYANtoU_MATRIX(t3d);

     case t3dMixedType(T3D_U_BRYAN, T3D_U_VECTOR):
       t3dConvertU_BRYANtoU_MATRIX(t3d);
       return t3dConvertU_MATRIXtoU_VECTOR(t3d);

     /* From VECTOR */
     case t3dMixedType(T3D_VECTOR, T3D_MATRIX):
       return t3dConvertVECTORtoMATRIX(t3d);

     case t3dMixedType(T3D_VECTOR, T3D_BRYAN):
       t3dConvertVECTORtoMATRIX(t3d);
       return t3dConvertMATRIXtoBRYAN(t3d);

     case t3dMixedType(T3D_U_VECTOR, T3D_U_MATRIX):
       return t3dConvertU_VECTORtoU_MATRIX(t3d);

     case t3dMixedType(T3D_U_VECTOR, T3D_U_BRYAN):
       t3dConvertU_VECTORtoU_MATRIX(t3d);
       return t3dConvertU_MATRIXtoU_BRYAN(t3d);
   }

   return NULL;
}


/* --- r3dConvertTo -------------------------------------------------- */

/** \brief Convert R3D to the "typeTo" representation.
 *
 * \param[in] typeTo	Type to convert to.
 * \param[in] r3d	R3D object to convert.
 *
 * \return r3d, converted.
 */

R3D *
r3dConvertTo(T3DType typeTo, R3D *r3d)
{
  return (R3D *)t3dConvertTo(typeTo, (T3D *)r3d);
}


/* --- r3dConvertToT3D ----------------------------------------------- */

/** \brief Convert R3d to a generic T3D.
 *
 * \param[in] r3d	R3D object to convert.
 *
 * \return r3d, converted.
 */

T3D *
r3dConvertToT3D(R3D *r3d)
{
  return (T3D *)r3d;
}


/* --- t3dConvertToU_ ------------------------------------------------ */

/* Conversion from certain to uncertain types */

static T3D *
t3dConvertToU_(T3D *t3d)
{
  if (t3dUncertain(t3d->type)) return t3d;

  switch(t3d->type) {
    case T3D_MATRIX:
      memset(t3d->umatrix.v, 0, sizeof(t3d->umatrix.v));
      t3d->type = T3D_U_MATRIX;
      break;

    case T3D_BRYAN:
      memset(t3d->ubryan.v, 0, sizeof(t3d->ubryan.v));
      t3d->type = T3D_U_BRYAN;
      break;

    case T3D_VECTOR:
      memset(t3d->uvector.v, 0, sizeof(t3d->uvector.v));
      t3d->type = T3D_U_VECTOR;
      break;
  }
  return t3d;
}


/* --- t3dConvertToC_ ------------------------------------------------ */

/* Conversion from uncertain to certain types */

static T3D *
t3dConvertToC_(T3D *t3d)
{
  if (!t3dUncertain(t3d->type)) return t3d;

  switch(t3d->type) {
    case T3D_U_MATRIX:	t3d->type = T3D_MATRIX;	break;
    case T3D_U_BRYAN:	t3d->type = T3D_BRYAN;	break;
    case T3D_U_VECTOR:	t3d->type = T3D_VECTOR;	break;
  }
  return t3d;
}


/* --- t3dConvertMATRIXtoBRYAN ----------------------------------------- */

/** \brief Conversion from T3D_MATRIX to T3D_BRYAN
 *
 * \param[in] t3d	Object to convert.
 * \retun the converted transformation.
 */

static T3D *
t3dConvertMATRIXtoBRYAN(T3D *t3d)
{
  T3DBryan c;
  if (t3d->flags & T3D_MINIMAL_SIZE) return NULL;

  /* compute Euler parameters */
  t3dEulerOfMatrix(t3d->matrix.matrix, c.euler);

  /* update t3d */
  c.type = T3D_BRYAN;
  c.flags = t3d->flags;
  t3dSetOk((T3D *)&c);
  t3dCopyIn(t3d, (T3D *)&c);
  return t3d;
}


/* --- t3dConvertMATRIXtoVECTOR -------------------------------------- */

/* Conversion from T3D_MATRIX to T3D_VECTOR */

static T3D *
t3dConvertMATRIXtoVECTOR(T3D *t3d)
{
  T3DVector c;
  if (t3d->flags & T3D_MINIMAL_SIZE) return NULL;

  /* compute Vector parameters */
  t3dVectorOfMatrix(t3d->matrix.matrix, c.vector);

  /* update t3d */
  c.type = T3D_VECTOR;
  c.flags = t3d->flags;
  t3dSetOk((T3D *)&c);
  t3dCopyIn(t3d, (T3D *)&c);
  return t3d;
}


/* --- t3dConvertU_MATRIXtoU_BRYAN ----------------------------------- */

/* Conversion from T3D_MATRIX to T3D_U_BRYAN */

static T3D *
t3dConvertU_MATRIXtoU_BRYAN(T3D *t3d)
{
  T3DUBryan c;
  if (t3d->flags & T3D_MINIMAL_SIZE) return NULL;

  /* compute Euler parameters and copy covariances */
  t3dEulerOfMatrix(t3d->umatrix.matrix, c.euler);
  memcpy(c.v, t3d->umatrix.v, sizeof(c.v));

  /* update t3d */
  c.type = T3D_U_BRYAN;
  c.flags = t3d->flags;
  t3dSetOk((T3D *)&c);
  t3dCopyIn(t3d, (T3D *)&c);
  return t3d;
}


/* --- t3dConvertU_MATRIXtoU_VECTOR ---------------------------------- */

/* Conversion from T3D_U_MATRIX to T3D_U_VECTOR */

static T3D *
t3dConvertU_MATRIXtoU_VECTOR(T3D *t3d)
{
  double *m;
  double euler[3];

  double JvR[27], JRea[36], Jvea[9];
  double JRa[9], JRb[9], JRg[9];

  double dtst, co, t, f;

  T3DUVector c;
  if (t3d->flags & T3D_MINIMAL_SIZE) return NULL;

  /* compute Vector parameters */
  t3dVectorOfMatrix(t3d->umatrix.matrix, c.vector);

  /* covariances */
  m = t3d->umatrix.matrix;
  co = (m[0] + m[5] + m[10] - 1.)/2.0;
  if (co > 1.0) co = 1.0; else if (co < -1.0) co = -1.0;
  t = acos(co);

  /* Drives termes vecteur de rotation / termes de la matrice */
  if (fabs(t)< 1e-10) {
    /* theta almost 0, use the limit */
    f=1.0;
    dtst=-1.0/6.0;
  } else {
    if (1+co < 1e-10) {
      /* theta almost PI */
      f=1.0;
      dtst=1.0;
    } else {
      double dt;

      f = t/sin(t);
      dt= -0.5/(sqrt(1.0-(co*co)));
      dtst= dt*(sin(t)-t*cos(t))/(sin(t)*sin(t));
    }
  }

  t3d_dm_fill(
    JvR, 3, 9,
    dtst*(m[9] - m[6])/2.0, 0.0,                    0.0,
    0.0,                    dtst*(m[9] - m[6])/2.0, -0.5*f,
    0.0,                    0.5*f,                  dtst*(m[9] - m[6])/2.0,

    dtst*(m[2] - m[8])/2.0, 0.0,                    0.5*f,
    0.0,                    dtst*(m[2] - m[8])/2.0, 0.0,
    -0.5*f,                 0.0,                    dtst*(m[2] - m[8])/2.0,

    dtst*(m[4] - m[1])/2.0, -0.5*f,                 0.0,
    0.5*f,                  dtst*(m[4] - m[1])/2.0, 0.0,
    0.0,                    0.0,                    dtst*(m[4] - m[1])/2.0
    );

  /* Drives de la matrice de rotation par raport aux angles d'Euler */
  t3dEulerAnglesOfMatrix(m, euler);
  t3dDerivativeOfEuler(euler, JRa, JRb, JRg);
  t3d_dm_blocfill(JRea, 9, 3, 9, 1, JRa, JRb, JRg);

  /* The resulting jacobian matrix */
  t3d_dm_mult(JvR, JRea, Jvea, 3, 9, 9, 3);
  t3d_dm_blocfill(JRea, 6, 6, 3, 3, Jvea, NULL, NULL, t3dmid3);
  t3d_dmfsf_jmult_6x6(JRea, t3d->umatrix.v, c.v);

  /* update t3d */
  c.type = T3D_U_VECTOR;
  c.flags = t3d->flags;
  t3dSetOk((T3D *)&c);
  t3dCopyIn(t3d, (T3D *)&c);
  return t3d;
}


/* --- t3dConvertBRYANtoMATRIX --------------------------------------- */

/* Conversion from T3D_BRYAN to T3D_MATRIX */

static T3D *
t3dConvertBRYANtoMATRIX(T3D *t3d)
{
  T3DMatrix c;
  if (t3d->flags & T3D_MINIMAL_SIZE) return NULL;

  /* compute Bryan parameters */
  t3dMatrixOfEuler(t3d->bryan.euler, c.matrix);

  /* update t3d */
  c.type = T3D_MATRIX;
  c.flags = t3d->flags;
  t3dSetOk((T3D *)&c);
  t3dCopyIn(t3d, (T3D *)&c);
  return t3d;
}


/* --- t3dConvertU_BRYANtoU_MATRIX ----------------------------------- */

/* Conversion from T3D_U_BRYAN to T3D_U_MATRIX */

static T3D *
t3dConvertU_BRYANtoU_MATRIX(T3D *t3d)
{
  T3DUMatrix c;
  if (t3d->flags & T3D_MINIMAL_SIZE) return NULL;

  /* compute Bryan parameters */
  t3dMatrixOfEuler(t3d->ubryan.euler, c.matrix);
  memcpy(c.v, t3d->ubryan.v, sizeof(c.v));

  /* update t3d */
  c.type = T3D_U_MATRIX;
  c.flags = t3d->flags;
  t3dSetOk((T3D *)&c);
  t3dCopyIn(t3d, (T3D *)&c);
  return t3d;
}


/* --- t3dConvertVECTORtoMATRIX -------------------------------------- */

/* Conversion from T3D_VECTOR to T3D_MATRIX */

static T3D *
t3dConvertVECTORtoMATRIX(T3D *t3d)
{
  T3DMatrix c;
  if (t3d->flags & T3D_MINIMAL_SIZE) return NULL;

  /* compute Bryan parameters */
  t3dMatrixOfVector(t3d->vector.vector, c.matrix);

  /* update t3d */
  c.type = T3D_MATRIX;
  c.flags = t3d->flags;
  t3dSetOk((T3D *)&c);
  t3dCopyIn(t3d, (T3D *)&c);
  return t3d;
}


/* --- t3dConvertU_VECTORtoU_MATRIX ---------------------------------- */

/* Conversion from T3D_U_VECTOR to T3D_U_MATRIX */

static T3D *
t3dConvertU_VECTORtoU_MATRIX(T3D *t3d)
{
  double *m;
  double dRx[9], dRy[9], dRz[9];
  double JeaR[27], Jeav[9], JRv[36];
  double s, d, d2;

  T3DUMatrix c;
  if (t3d->flags & T3D_MINIMAL_SIZE) return NULL;

  /* compute Bryan parameters */
  t3dMatrixOfVector(t3d->vector.vector, c.matrix);

  /* Drives angles d'Euler / termes de la matrice */
  m = c.matrix;
  d2 = m[0]*m[0] + m[4]*m[4];
  d = sqrt(d2);
  s = m[10]*m[10]+m[9]*m[9];

  t3d_dm_fill(
    JeaR, 3, 9,
    -m[4]/d2,    0.0, 0.0, m[0]/d2,     0.0, 0.0, 0.0, 0.0,     0.0,
    m[8]*m[0]/d, 0.0, 0.0, m[8]*m[4]/d, 0.0, 0.0, -d,  0.0,     0.0,
    0.0,             0.0, 0.0, 0.0,     0.0, 0.0, 0.0, m[10]/s, -m[9]/s);

  /* Drives de la matrice de rotation par raport au vecteur rotation */
  t3dDerivativeOfVector(t3d->uvector.vector, dRx, dRy, dRz);
  t3d_dm_blocfill(JRv, 9, 3, 9, 1, dRx, dRy, dRz);

  /* The resulting jacobian matrix */
  t3d_dm_mult(JeaR, JRv, Jeav, 3, 9, 9, 3);
  t3d_dm_blocfill(JRv, 6, 6, 3, 3, Jeav, NULL, NULL, t3dmid3);
  t3d_dmfsf_jmult_6x6(JRv, t3d->uvector.v, c.v);

  /* update t3d */
  c.type = T3D_U_MATRIX;
  c.flags = t3d->flags;
  t3dSetOk((T3D *)&c);
  t3dCopyIn(t3d, (T3D *)&c);
  return t3d;
}
