/*	$LAAS: t3deuler.c,v 1.1 2005/08/31 21:56:52 sara Exp $ */

/* 
 * Copyright (c) 2000-2004 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <math.h>

#include "t3d/t3d.h"
#include "t3d_internal.h"


/*
 * --- t3dEulerAnglesOfMatrix -------------------------------------------
 *
 * Compute euler angles from homogeneous matrix, put them in euler[0..2].
 */

void
t3dEulerAnglesOfMatrix(double *mat, double *euler)
{
   double d;

   d = sqrt(mat[0]*mat[0] + mat[4]*mat[4]);
   
   if (fabs(d) > 1e-10) {
      euler[0] = atan2(mat[4], mat[0]);
      euler[2] = atan2(mat[9], mat[10]);
   } else {
      euler[0] = atan2(-mat[1], mat[5]);
      euler[2] = 0.0;
   }
   euler[1] = atan2(-mat[8], d);
}

/*
 * --- t3dEulerOfMatrix -------------------------------------------------
 *
 * Compute euler angles and translation from homogeneous matrix, put them
 * in euler[0..5].
 */

void
t3dEulerOfMatrix(double *mat, double *euler)
{
   t3dEulerAnglesOfMatrix(mat, euler);
  
   euler[3] = mat[3];
   euler[4] = mat[7];
   euler[5] = mat[11];
}


/*
 * --- t3dDerivativeOfEuler ---------------------------------------------
 * (t3dDerivativeOfEuler was previously t3deul.c:t3deul_rdr)
 *
 * Compute derivative rotation matrices of Euler angles
 */

void
t3dDerivativeOfEuler(double *euler, double *dRa, double *dRb, double *dRg)
{
   double sa, ca, cb, sb, cg, sg;

   ca = cos(euler[0]); sa = sin(euler[0]);
   cb = cos(euler[1]); sb = sin(euler[1]);
   cg = cos(euler[2]); sg = sin(euler[2]);

   /* derivative /euler[0] */
   dRa[0] = -sa*cb;
   dRa[1] = -sa*sb*sg - ca*cg;
   dRa[2] = -sa*sb*cg + ca*sg;
   dRa[3] = ca*cb;
   dRa[4] = ca*sb*sg - sa*cg;
   dRa[5] = ca*sb*cg + sa*sg;
   dRa[6] = 0.0;
   dRa[7] = 0.0;
   dRa[8] = 0.0;
  
   /* derivative /euler[1] */
   dRb[0] = - ca*sb;
   dRb[1] = ca*cb*sg;
   dRb[2] = ca*cb*cg;
   dRb[3] = -sa*sb;
   dRb[4] = sa*cb*sg;
   dRb[5] = sa*cb*cg;
   dRb[6] = -cb;
   dRb[7] = -sb*sg;
   dRb[8] = -sb*cg;

   /* derivative /euler[2] */
   dRg[0] = 0.0;
   dRg[1] = ca*sb*cg + sa*sg;
   dRg[2] = -ca*sb*sg + sa*cg;
   dRg[3] = 0.0;
   dRg[4] = sa*sb*cg - ca*sg;
   dRg[5] = -sa*sb*sg - ca*cg;
   dRg[6] = 0.0;
   dRg[7] = cb*cg;
   dRg[8] = -cb*sg;
}
