/*	$LAAS: t3dinvert.c,v 1.2 2005/08/31 21:56:52 sara Exp $ */

/*
 * Copyright (c) 2000-2004 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "t3d/t3d.h"
#include "t3d_internal.h"


/* Local functions --------------------------------------------------- */

static void	invertMatrix(T3D *t3d);
static void	invertVector(T3D *t3d);


/* --- t3dInvert ----------------------------------------------------- */

/** \brief Invert a T3D.
 *
 * The T3D might be converted to another representation for doing this,
 * and since not all conversion are possible, you might want to check
 * errno.
 * Note that the T3D is overwritten.
 *
 * \param[in] t3d	T3D object.
 *
 * \return t3d, inverted.
 */

T3D *
t3dInvert(T3D *t3d)
{
   if (!t3dOk(t3d)) return NULL;

   switch(t3d->type) {
     case T3D_MATRIX:
     case T3D_U_MATRIX:
       invertMatrix(t3d);
       break;

     case T3D_VECTOR:
     case T3D_U_VECTOR:
       invertVector(t3d);
       break;

     case T3D_BRYAN:
     case T3D_U_BRYAN:
       if (t3d->flags & T3D_ALLOW_CONVERSION) {
	 if (t3dUncertain(t3d->type))
	   t3dConvertTo(T3D_U_MATRIX, t3d);
	 else
	   t3dConvertTo(T3D_MATRIX, t3d);

	 invertMatrix(t3d);
       } else
	 t3dSetError(t3d, T3D_ENOTIMPLEMENTED);
       break;
   }

   return t3dOk(t3d)? t3d : NULL;
}


/* --- t3dInvertIn --------------------------------------------------- */

/** \brief Invert T3D.
 *
 * Invert T3D `a', store the result in c and return c. See t3Invert().
 *
 * \param[in] t3d	T3D object.
 * \param[in] inv	Where to store the inverse.
 *
 * \return inv, equel to the inverse of t3d.
 */

T3D *
t3dInvertIn(T3D *inv, T3D *t3d)
{
  return t3dInvert(t3dCopyIn(inv, t3d));
}


/* --- r3dInvert ----------------------------------------------------- */

/** \brief Invert a R3D.
 *
 * The R3D might be converted to another representation for doing this,
 * and since not all conversion are possible, you might want to check
 * errno.
 * Note that the R3D is overwritten.
 *
 * \param[in] r3d	R3D object.
 *
 * \return r3d, inverted.
 */

R3D *
r3dInvert(R3D *r3d)
{
  return (R3D *)t3dInvert((T3D *)r3d);
}


/*
 * --- invertMatrix -----------------------------------------------------
 *
 * Invert a matrix T3D.
 */

static void
invertMatrix(T3D *t3d)
{
   static const double JRtR[81] = {
      1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
      0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0,
      0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0,
      0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
      0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0,
      0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0,
      0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
      0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0,
      0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0
   };

   double *m = t3d->matrix.matrix;

   double i[16], cov[21];
   double euler[3], J[36], JeRt[27], Jee[9];
   double dra[9], drb[9], drg[9], r2[9];
   double j1[3], j2[3], j3[3];
   double t[3];
   double tmp[27];
   double s, d, d2;

   /*
    * matrix inversion
    */

   i[0] = m[0];
   i[1] = m[4];
   i[2] = m[8];
   i[3] = -(m[3]*m[0]+ m[7]*m[4]+ m[11]*m[8]);

   i[4] = m[1];
   i[5] = m[5];
   i[6] = m[9];
   i[7] = -(m[3]*m[1]+ m[7]*m[5]+ m[11]*m[9]);

   i[8] = m[2];
   i[9] = m[6];
   i[10]= m[10];
   i[11] = -(m[3]*m[2]+ m[7]*m[6]+ m[11]*m[10]);

   i[12] = 0.0;
   i[13] = 0.0;
   i[14] = 0.0;
   i[15] = 1.0;

   if (!t3dUncertain(t3d->type)) goto done_certain;

   /*
    * Variances
    */


   /* Drives angles d'Euler / termes de la matrice */
   d = sqrt(d2 = (i[0]*i[0] + i[4]*i[4]));
   s = i[10]*i[10]+i[9]*i[9];
   t3d_dm_fill(
      JeRt, 3, 9,
      -i[4]/d2,    0.0, 0.0, i[0]/d2,     0.0, 0.0, 0.0, 0.0,     0.0,
      i[8]*i[0]/d, 0.0, 0.0, i[8]*i[4]/d, 0.0, 0.0, -d,  0.0,     0.0,
      0.0,         0.0, 0.0, 0.0,         0.0, 0.0, 0.0, i[10]/s, -i[9]/s
      );

   /* Jacobienne de la matrice de rotation */
   t3dEulerAnglesOfMatrix(m, euler);
   t3dDerivativeOfEuler(euler, dra, drb, drg);
   t3d_dm_blocfill(J, 9, 3, 9, 1, dra, drb, drg);

   /* We deduce the jacobian of the inversion matrix */
   t3d_dm_mult(JRtR, J, tmp, 9, 9, 9, 3);
   t3d_dm_mult(JeRt, tmp, Jee, 3, 9, 9, 3);

   t[0] = -m[3]; t[1] = -m[7]; t[2] = -m[11];
   t3d_dm_trans_3x3(dra); t3d_dm_mult(dra, t, j1, 3, 3, 3, 1);
   t3d_dm_trans_3x3(drb); t3d_dm_mult(drb, t, j2, 3, 3, 3, 1);
   t3d_dm_trans_3x3(drg); t3d_dm_mult(drg, t, j3, 3, 3, 3, 1);
   t3d_dm_blocfill(tmp, 3, 3, 3, 1, j1, j2, j3);

   t3d_dm_fill(r2, 3, 3,
	       -i[0], -i[1], -i[2],
	       -i[4], -i[5], -i[6],
	       -i[8], -i[9], -i[10]);
   t3d_dm_blocfill(J, 6, 6, 3, 3, Jee, NULL, tmp, r2);

   /* And the propagation through the inversion operation */
   t3d_dmfsf_jmult_6x6(J, t3d->umatrix.v, cov);

   /* update t3d */
   memcpy(t3d->umatrix.v, cov, sizeof(cov));
done_certain:
   memcpy(t3d->umatrix.matrix, i, sizeof(i));
}


/*
 * --- invertVector -----------------------------------------------------
 *
 * Invert vector T3D.
 */

static void
invertVector(T3D *t3d)
{
   double *v = t3d->vector.vector;

   double rot[9], tr[3], cov[21];
   double drx[9], dry[9], drz[9];
   double J[36];

   double t, t2, f, c, g;

   /* Rotation vector is just taken opposed */
   v[0] = -v[0]; v[1] = -v[1]; v[2] = -v[2];

   /* Compute translation */
   t = sqrt(t2 = (v[0]*v[0] + v[1]*v[1] + v[2]*v[2]));
   if ( t > 1e-10 ) {
      f = sin(t)/t;
      c = cos(t);
      g = (1-c)/t2;

      t3d_dm_fill(
	 rot, 3, 3,
	 c + v[0]*v[0] * g,    v[0]*v[1] * g - v[2]*f, v[0]*v[2] * g + v[1]*f,
	 v[0]*v[1] * g + v[2]*f, c + v[1]*v[1] * g,    v[1]*v[2] * g - v[0]*f,
	 v[0]*v[2] * g - v[1]*f, v[1]*v[2] * g + v[0]*f, c + v[2]*v[2] * g );
   } else /* theta = 0, identity transform */
      memcpy(rot, t3dmid3, sizeof(rot));

   tr[0] = - rot[0] * v[3] - rot[1] * v[4] - rot[2] * v[5];
   tr[1] = - rot[3] * v[3] - rot[4] * v[4] - rot[5] * v[5];
   tr[2] = - rot[6] * v[3] - rot[7] * v[4] - rot[8] * v[5];

   if (!t3dUncertain(t3d->type)) goto done_certain;

   /*
    * Variances
    */

   t3dDerivativeOfVector(v, drx, dry, drz);

   /* jacobian matrix of the inversion */
   t3d_dm_fill(
      J, 6, 6,

      -1.0, 0.0, 0.0, 0.0, 0.0, 0.0,
      0.0, -1.0, 0.0, 0.0, 0.0, 0.0,
      0.0, 0.0, -1.0, 0.0, 0.0, 0.0,
      drx[0] * v[3] + drx[1] * v[4] + drx[2] * v[5],
      dry[0] * v[3] + dry[1] * v[4] + dry[2] * v[5],
      drz[0] * v[3] + drz[1] * v[4] + drz[2] * v[5],
      -rot[0], -rot[1], -rot[2],
      drx[3] * v[3] + drx[4] * v[4] + drx[5] * v[5],
      dry[3] * v[3] + dry[4] * v[4] + dry[5] * v[5],
      drz[3] * v[3] + drz[4] * v[4] + drz[5] * v[5],
      -rot[3], -rot[4], -rot[5],
      drx[6] * v[3] + drx[7] * v[4] + drx[8] * v[5],
      dry[6] * v[3] + dry[7] * v[4] + dry[8] * v[5],
      drz[6] * v[3] + drz[7] * v[4] + drz[8] * v[5],
      -rot[6], -rot[7], -rot[8]
      );

   /* propagation through the inversion operation */
   t3d_dmfsf_jmult_6x6(J, t3d->uvector.v, cov);

   /* update t3d */
   memcpy(t3d->uvector.v, cov, sizeof(cov));
done_certain:
   t3d->uvector.vector[3] = tr[0];
   t3d->uvector.vector[4] = tr[1];
   t3d->uvector.vector[5] = tr[2];
}
