/* 
 * Copyright (c) 2007,2009 Anthony Mallet                  --  Mon Feb  5 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdio.h>
#include <string.h>

#include "t3d/t3d.h"
#include "t3d/r3d.h"
#include "t3d/v3d.h"
#include "t3d_internal.h"

#ifdef WIN32
#define snprintf _snprintf
#endif /*WIN32*/

/* --- Local data ---------------------------------------------------- */

static int	t3d_fwrite(void *cookie, const char *buf, int len);


/* --- t3dFPrint ----------------------------------------------------- */

/** \brief Output a ASCII T3D description on a stream
 *
 * \param[in] stream	Stream to dump on.
 * \param[in] t3d	T3D to dump.
 *
 * \return the number of byte written.
 */

int
t3dFPrint(FILE *stream, const T3D *t3d)
{
  return t3dSPrint(t3d, t3d_fwrite, stream);
}


/* --- r3dFPrint ----------------------------------------------------- */

/** \brief Output a ASCII R3D description on a stream
 *
 * \param[in] stream	Stream to dump on.
 * \param[in] r3d	R3D to dump.
 *
 * \return the number of byte written.
 */

int
r3dFPrint(FILE *stream, const R3D *r3d)
{
  return t3dSPrint((T3D *)r3d, t3d_fwrite, stream);
}


/* --- v3dFPrint ----------------------------------------------------- */

/** \brief Output a ASCII V3D description on a stream
 *
 * \param[in] stream	Stream to dump on.
 * \param[in] v3d	V3D to dump.
 *
 * \return the number of byte written.
 */

int
v3dFPrint(FILE *stream, const V3D *v3d)
{
  return v3dSPrint(v3d, t3d_fwrite, stream);
}


/* --- t3d_fwrite ---------------------------------------------------- */

static int
t3d_fwrite(void *cookie, const char *buf, int len)
{
  return fwrite(buf, len, 1, (FILE *)cookie);
}


/* --- t3dSPrint ----------------------------------------------------- */

/** \brief Output a ASCII T3D to user-functions
 *
 * \param[in] t3d	T3D to dump.
 * \param[in] write	Write function.
 * \param[in] cookie	Argument passed to write function.
 *
 * \return the number of bytes written.
 */

int
t3dSPrint(const T3D *t3d, int (*write)(void *, const char *, int), void *cookie)
{
  char buf[128];
  int i, j, l;
  int n = 0;

  switch(t3d->type) {
    case T3D_MATRIX:
    case T3D_U_MATRIX:
      if (t3dUncertain(t3d->type))
	n += write(cookie, "T3D_U_MATRIX\n", 13);
      else
	n += write(cookie, "T3D_MATRIX\n", 11);

      for(i=0;i<16;i+=4) {
	n += write(cookie, "| ", 2);
	for(j=0; j<4; j++) {
	  l = snprintf(buf, sizeof(buf),
		       "%12g%s", t3d->matrix.matrix[i+j], j<3?"\t":"");
	  n += write(cookie, buf, l);
	}
	n += write(cookie, " |\n", strlen(" |\n"));
      }
      break;

    case T3D_BRYAN:
    case T3D_U_BRYAN:
      l = snprintf(buf, sizeof(buf), "| x: %12g\t", t3d->bryan.euler[3]);
      n += write(cookie, buf, l);
      l = snprintf(buf, sizeof(buf), "y: %12g\t\t", t3d->bryan.euler[4]);
      n += write(cookie, buf, l);
      l = snprintf(buf, sizeof(buf), "z: %12g |\n", t3d->bryan.euler[5]);
      n += write(cookie, buf, l);

      l = snprintf(buf, sizeof(buf), "| yaw: %12g\t", t3d->bryan.euler[0]);
      n += write(cookie, buf, l);
      l = snprintf(buf, sizeof(buf), "pitch: %12g\t", t3d->bryan.euler[1]);
      n += write(cookie, buf, l);
      l = snprintf(buf, sizeof(buf), "roll: %12g |\n", t3d->bryan.euler[2]);
      n += write(cookie, buf, l);
      break;
  }
  return n;
}


/* --- v3dSPrint ----------------------------------------------------- */

/** \brief Output a ASCII V3D to user-functions
 *
 * \param[in] v3d	V3D to dump.
 * \param[in] write	Write function.
 * \param[in] cookie	Argument passed to write function.
 *
 * \return the number of bytes written.
 */

int
v3dSPrint(const V3D *v3d, int (*write)(void *, const char *, int), void *cookie)
{
  char buf[128];
  int i, l;
  int n = 0;

  n += write(cookie, "| ", 2);
  for(i=0; i<4; i++) {
    l = snprintf(buf, sizeof(buf), "%12g%s", v3d->v[i],i<3?"\t":"");
    n += write(cookie, buf, l);
  }
  n += write(cookie, " |\n", strlen(" |\n"));

  return n;
}
