/*	$LAAS: t3dmatrix.c,v 1.1 2005/08/31 21:56:52 sara Exp $ */

/*
 * Copyright (c) 1990,2004 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

/* ATTENTION (for LAAS users):
 *
 * This file contains functions that are in libmatrices from
 * calife, because there is not (yet) an official release of this
 * library. All the dm_* functions have been prefixed by t3d_dm_*.
 *
 * If a package containing libmatrices is ever released, the file matrix.c
 * in this library (libt3d) should be removed, the t3d_dm prefix replaced by
 * dm, and a dependency on libmatrices created.
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>

#include "t3d/t3d.h"
#include "t3d_internal.h"

/* ----------------------------------------------------------------------
 *
 * Operations on matrices.
 *
 * --- Full matrices:
 *
 *  [ 1 2 3 ]
 *  [ 4 5 6 ]
 *  [ 7 8 9 ]
 *
 *  is stored as [1 2 3 4 5 6 7 8 9].
 *
 * --- Symmetric matrices:
 *
 *  [ 1 2 4 ]
 *  [ 2 3 5 ]
 *  [ 4 5 6 ]
 *
 *  is stored as [1 2 3 4 5 6].
 *
 * --- Operations:
 *
 * Full matrices
 *  t3d_dm_fill(m, l, c, coeff1, coeff2, ...)		  fill m by values
 *  t3d_dm_blocfill(mat, l, c, lb, cb, bloc1, ..., blocn) fill m by blocs
 *  t3d_dm_mulscal(m1, scalaire, m, l, c)		  m = scalaire . m1
 *  t3d_dm_trans_3x3(m)					  m = m_transposed
 *  t3d_dm_mult(m1, m2, m, l1, c1, l2, c2)		  m = m1 * m2
 *
 * Symmetric matrices
 *  t3d_dms_plus_6x6(m1, m2, m)				  m = m1 + m2
 *  t3d_dmfsf_jmult_6x6(m1, m2, m)		m = m1 . m2. m1_transposed
 */


/* --- local data ---------------------------------------------------- */

#define MATELT(mat, i, j, c)	(*((mat) + (c) * (i) + (j)) )

#define INDEX_SMATELT(i, j)   \
   ((i)<(j) ? ((j)*((j)+1))/2 + (i) : ((i)*((i)+1))/2 +(j))
#define SMATELT(mat, i, j, n)	(*((mat) + INDEX_SMATELT(i,j)))

/* 3x3 indentity matrix */
const double t3dmid3[9] = {
   1.0, 0.0, 0.0,
   0.0, 1.0, 0.0,
   0.0, 0.0, 1.0
};


/* --- t3d_dm_fill ------------------------------------------------------
 *
 * fill matrix m by values
 *
 *	with: m full		lxc
 *	      val1, ...		double
 *
 * Example : t3d_dm_fill(mat, 2, 3, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0)
 *
 *     m = [ 1.0 2.0 3.0 ]
 *	   [ 4.0 5.0 6.0 ]
 */

int
t3d_dm_fill(double *mat, int l, int c, ...)
{
   va_list ap;
   int i;

   va_start(ap, c);

   for (i = 0; i < l*c; i++)
      *mat++ = va_arg (ap, double);

   va_end (ap);
   return 0;
}


/* --- t3d_dm_blocfill --------------------------------------------------
 *
 * fill matrix m by blocs
 *
 *	with: m full		lxc
 *	      bloc1, ... full	lb x lc
 *
 * if bloci == NULL, fill with zeroes.
 *
 * Example:	bloc1 = [1.0 2.0]
 *		bloc2 = [3.0 4.0]
 *		bloc3 = NULL
 *		bloc4 = [7.0 8.0]
 *
 * t3d_dm_blocfill(m, 2, 4, 1, 2, bloc1, bloc2, bloc3, bloc4)
 *
 *     m = [ bloc1 ! bloc2 ]   [ 1.0 2.0 3.0 4.0 ]
 *         [ bloc3 ! bloc4 ]   [ 0.0 0.0 7.0 8.0 ]
 */

int
t3d_dm_blocfill(double *m, int l, int c, int lb, int cb, ...)
{
   int i, j, ib, jb;
   double *bloc;
   va_list ap;

   if (l%lb != 0 || c%cb != 0) {
     fprintf(stderr, "invalid call of t3d_dm_blocfill\n");
     return 1;
   }

   va_start(ap, cb);

   for (ib = 0; ib < l/lb; ib++) {
      for (jb = 0; jb < c/cb; jb++) {
	 bloc = va_arg(ap, double *);
	 if (bloc == NULL) {
	    for (i = ib * lb; i < (ib + 1) * lb; i++)
	       for (j = jb * cb; j < (jb + 1) * cb; j++)
		  *(m + c * i + j) = 0.0;
	 } else {
	    for (i = ib * lb; i < (ib + 1) * lb; i++)
	       for (j = jb * cb; j < (jb + 1) * cb; j++)
		  *(m + c * i + j) = *bloc++;
	 }
      }
   }

   va_end(ap);
   return 0;
}


/* --- t3d_dm_dm_mulscal ------------------------------------------------
 *
 * produc     m = scal * m1
 *
 *	with : m full		lxc
 *
 * Result is full, lxc.
 */

int
t3d_dm_mulscal(double *m1, double scal, double *m, int l, int c)
{
   int k;

   for (k = 0; k < l*c; k++)
      *m++ = scal * *m1++;

   return 0;
}


/* --- t3d_dm_mult ------------------------------------------------------
 *
 * product     m1 * m2
 *
 *	with : m1 full		l1xc1
 *	       m2 full		l2xc2
 *
 * Result is full, l1xc2.
 */

int
t3d_dm_mult(const double *m1, const double *m2, double *m,
	    int l1, int c1, int l2, int c2)
{
   int i, j, k;
   double val;

   if (c1 != l2) {
     fprintf(stderr, "invalid call of t3d_dm_mult\n");
     return 1;
   }
   if (m == m1 || m == m2) {
     fprintf(stderr, "invalid call of t3d_dm_mult\n");
     return 1;
   }

   for (i = 0; i < l1; i++)
      for (j = 0; j < c2; j++) {
	 val = 0.;
	 for (k = 0; k < c1; k++)
	    val += MATELT(m1, i, k, c1) * MATELT(m2, k, j, c2);
	 *m++ = val;
      }

   return 0;
}


/* --- t3d_dm_trans_3x3 -------------------------------------------------
 *
 * transpose     m
 *
 *	with : m full		3x3
 *
 * Result is full, 3x3.
 */

int
t3d_dm_trans_3x3(double *m)
{
   double tmp;

   tmp = m[1]; m[1] = m[3]; m[3] = tmp;
   tmp = m[2]; m[2] = m[6]; m[6] = tmp;
   tmp = m[5]; m[5] = m[7]; m[7] = tmp;
   return 0;
}


/* --- t3d_dms_plus_6x6 -------------------------------------------------
 *
 * sum		m1 + m2
 *
 *	with:	m1 symmetric	6x6
 *		m2 symmetric	6x6
 *
 * Result is symmetric
 */

int
t3d_dms_plus_6x6(double *m1, double *m2, double *m)
{
   int k;

   for (k = 0; k < 21; k++)
      *m++ = *m1++ + *m2++;
   return 0;
}


/* --- t3d_dmfsf_jmult_6x6 ----------------------------------------------
 *
 * product     m1 * m2 * m1_transposed
 *
 *	with : m1 full		6x6
 *	       m2 symmetric	6x6
 *
 * Result is always symmetric.
 *
 * t3d_dmfsf_jmult_6x6(m1, m2, m) ---> m = m1 * m2 * m1_transposed
 */

int
t3d_dmfsf_jmult_6x6(double *m1, double *m2, double *m)
{
   double aux[6*6], val;
   double *temp;
   int i, j, k;

   if (m == m1 || m == m2) {
     fprintf(stderr, "invalid call of t3d_dmfsf_jmult_6x6\n");
     return 1;
   }

   temp = aux;
   for (i = 0; i < 6; i++)
      for (j = 0; j < 6; j++) {
	 val = 0.;
	 for (k = 0; k < 6; k++)
	    val += MATELT(m1, i, k, 6) * SMATELT(m2, k, j, 6);
	*temp++ = val;
      }

   temp = aux;
   for (j = 0; j < 6; j++)
      for (i = 0; i <= j; i++) {
	 val = 0.;
	 for (k = 0; k < 6; k++)
	    val += MATELT(temp, i, k, 6) * MATELT(m1, j, k, 6);
	 *m++ = val;
      }

   return 0;
}


/*
 * --- t3dRotationOfMatrix ----------------------------------------------
 *
 * Get rotation 3x3 matrix from homogeneous matrix 4x4
 */

void
t3dRotationOfMatrix(double *m, double *r)
{
   r[0] = m[0];	   r[1] = m[1];	   r[2] = m[2];
   r[3] = m[4];	   r[4] = m[5];	   r[5] = m[6];
   r[6] = m[8];	   r[7] = m[9];	   r[8] = m[10];
}


/* --- t3dMatrixOfEuler ---------------------------------------------- */

/* Compute homogeneous matrix from Euler parameters. */

void
t3dMatrixOfEuler(double *euler, double *mat)
{
   double ca, sa, cb, sb, cg, sg;

   ca = cos(euler[0]); sa = sin(euler[0]);
   cb = cos(euler[1]); sb = sin(euler[1]);
   cg = cos(euler[2]); sg = sin(euler[2]);

   mat[0]  = ca*cb;
   mat[1]  = ca*sb*sg - sa*cg;
   mat[2]  = ca*sb*cg + sa*sg;
   mat[3]  = euler[3];

   mat[4]  = sa*cb;
   mat[5]  = sa*sb*sg + ca*cg;
   mat[6]  = sa*sb*cg - ca*sg;
   mat[7]  = euler[4];

   mat[8]  = -sb;
   mat[9]  = cb*sg;
   mat[10] = cb*cg;
   mat[11] = euler[5];

   mat[12] = 0.0;
   mat[13] = 0.0;
   mat[14] = 0.0;
   mat[15] = 1.0;
}

/* --- t3dMatrixOfVector --------------------------------------------- */

/* Compute homogeneous matrix from Vector parameters. */

void
t3dMatrixOfVector(double *vector, double *m)
{
  double rx = vector[0];
  double ry = vector[1];
  double rz = vector[2];
  double d;

  d = sqrt(rx*rx + ry*ry + rz*rz);
  /* if teta = 0, the matrix is set to identity */
  if ( d > 1e-10 ) {
    double f = sin(d) / d;
    double co = cos(d);
    double g = (1 - co) / d / d;

    t3d_dm_fill(
      m, 4, 4,
      co+rx*rx*g,   rx*ry*g-rz*f, rx*rz*g+ry*f, vector[3],
      rx*ry*g+rz*f, co+ry*ry*g,   ry*rz*g-rx*f, vector[4],
      rx*rz*g-ry*f, ry*rz*g+rx*f, co+rz*rz*g,   vector[5],
      0.0,          0.0,          0.0,          1.0);

  } else {
    t3d_dm_fill(
      m, 4, 4,
      1.0, 0.0, 0.0, vector[3],
      0.0, 1.0, 0.0, vector[4],
      0.0, 0.0, 1.0, vector[5],
      0.0, 0.0, 0.0, 1.0);
  }
}
