/* 
 * Copyright (c) 2007 Anthony Mallet                  --  Sun Mar  4 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */


#include <stdio.h>
#include <stdlib.h>

#include "t3d/r3d.h"
#include "t3d_internal.h"

#ifdef __cplusplus
namespace detail{
	const double id_4x4[16] = { 1.0, 0.0, 0.0,     0.0,
		0.0, 1.0, 0.0,     0.0,
		0.0, 0.0, 1.0,     0.0,
		0.0, 0.0, 0.0,     1.0 };

	const double zeros6[6] = {
		0.0, 0.0, 0.0,
		0.0, 0.0, 0.0
	};

	const double zeros21[21] = {
		0.0, 0.0, 0.0,	0.0, 0.0, 0.0,
		0.0, 0.0, 0.0,	0.0, 0.0, 0.0,
		0.0, 0.0, 0.0,	0.0, 0.0, 0.0,
		0.0, 0.0, 0.0
	};
}
const R3D r3dMatrixIdentity  (T3D_MATRIX,   0, T3D_EOK, detail::id_4x4);
const R3D r3dUMatrixIdentity (T3D_U_MATRIX, 0, T3D_EOK, detail::id_4x4, detail::zeros21);
const R3D r3dBryanIdentity   (T3D_BRYAN,    0, T3D_EOK, detail::zeros6);
const R3D r3dUBryanIdentity  (T3D_U_BRYAN,  0, T3D_EOK, detail::zeros6, detail::zeros21);
const R3D r3dVectorIdentity  (T3D_VECTOR,   0, T3D_EOK, detail::zeros6);
const R3D r3dUVectorIdentity (T3D_U_VECTOR, 0, T3D_EOK, detail::zeros6, detail::zeros21);

#else

const R3D r3dMatrixIdentity = { .matrix = {
    T3D_MATRIX, 0, T3D_EOK,
    { 1.0, 0.0, 0.0,     0.0,
      0.0, 1.0, 0.0,     0.0,
      0.0, 0.0, 1.0,     0.0,
      0.0, 0.0, 0.0,     1.0 }
  }};

const R3D r3dUMatrixIdentity = { .umatrix = {
    T3D_U_MATRIX, 0, T3D_EOK,
    { 1.0, 0.0, 0.0,     0.0,
      0.0, 1.0, 0.0,     0.0,
      0.0, 0.0, 1.0,     0.0,
      0.0, 0.0, 0.0,     1.0 },

    { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
           0.0, 0.0, 0.0, 0.0, 0.0,
                0.0, 0.0, 0.0, 0.0,
                     0.0, 0.0, 0.0,
                          0.0, 0.0,
                               0.0 }
  }};

const R3D r3dBryanIdentity = { .bryan = {
    T3D_BRYAN, 0, T3D_EOK,
    {{ 0.0, 0.0, 0.0,     0.0, 0.0, 0.0 }}
  }};

const R3D r3dUBryanIdentity = { .ubryan = {
    T3D_U_BRYAN, 0, T3D_EOK,
    {{ 0.0, 0.0, 0.0,     0.0, 0.0, 0.0 }},

    { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
           0.0, 0.0, 0.0, 0.0, 0.0,
                0.0, 0.0, 0.0, 0.0,
                     0.0, 0.0, 0.0,
                          0.0, 0.0,
                               0.0 }
  }};

const R3D r3dVectorIdentity = { .vector = {
    T3D_VECTOR, 0, T3D_EOK,
    {{ 0.0, 0.0, 0.0,     0.0, 0.0, 0.0 }}
  }};

const R3D r3dUVectorIdentity = { .uvector = {
    T3D_U_VECTOR, 0, T3D_EOK,
    {{ 0.0, 0.0, 0.0,     0.0, 0.0, 0.0 }},

    { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
           0.0, 0.0, 0.0, 0.0, 0.0,
                0.0, 0.0, 0.0, 0.0,
                     0.0, 0.0, 0.0,
                          0.0, 0.0,
                               0.0 }
  }};

#endif /*__cplusplus*/

/* --- r3dNew -------------------------------------------------------- */

/** \brief Malloc a new initialized structure for a R3D.
 *
 * Calls r3dInit() with an malloc'ed structure. Storage must be released
 * with r3dDelete().
 *
 * \param[in] type	Type of internal representation.
 * \param[in] flags	Flags for special behaviors.
 *
 * \return A pointer to the initialized R3D.
 */

R3D *
r3dNew(T3DType type, int flags)
{
  return (R3D *)t3dNew(type, flags);
}


/* --- r3dInit ------------------------------------------------------- */

/** \brief Initialize a R3D with the type "type".
 *
 * The size of the new structure is compatible with any type of
 * representation except if the T3D_MINIMAL_SIZE flag is present (in this
 * case the allocated size is minimal).
 *
 * \param[in] r3d	Pointer to an allocated R3D.
 * \param[in] type	Type of internal representation.
 * \param[in] flags	Flags for special behaviors.
 *
 * \return A pointer to the initialized R3D (equal to `r3d' parameter).
 */

R3D *
r3dInit(R3D *r3d, T3DType type, int flags)
{
  return (R3D *)t3dInit((T3D *)r3d, type, flags);
}


/* --- r3dDup -------------------------------------------------------- */

/** \brief Malloc a new structure for a R3D and initialize it with r3d
 *
 * \param[in] r3d	R3D object to duplicate.
 *
 * \return pointer on copied R3D.
 */

R3D *
r3dDup(R3D *r3d)
{
  if (!r3dOk(r3d)) { r3dWarn(r3d); return NULL; }
  return r3dCopyIn(r3dNew(r3d->type, r3d->flags), r3d);
}


/* --- r3dDelete ----------------------------------------------------- */

/** \brief Free an malloc'ed R3D.
 *
 * \param[in] r3d	R3D object.
 */

void
r3dDelete(R3D *r3d)
{
  t3dDelete((T3D *)r3d);
}
