/*	$Id: v3d.c 2008/01/21 12:05:44 mallet $ */

/*
 * Copyright (c) 2007 Anthony Mallet                  --  Thu Feb  1 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdlib.h>

#include <math.h>
#include "t3d/v3d.h"

/* --- v3dNew -------------------------------------------------------- */

/** \brief Malloc a new initialized structure for a V3D.
 *
 * Calls v3dInit() with an malloc'ed structure. Storage must be released
 * with v3dDelete().
 *
 * \return A pointer to the initialized V3D.
 */

V3D *
v3dNew()
{
   V3D *newV3D;

   newV3D = (V3D*) malloc(sizeof(V3D));
   if (newV3D == NULL) return NULL;

   return v3dInit(newV3D);
}


/* --- v3dInit ------------------------------------------------------- */

/** \brief Initialize a V3D.
 *
 * \param[in] v3d	Pointer to an allocated V3D.
 *
 * \return A pointer to the initialized V3D (equal to `v3d' parameter).
 */

V3D *
v3dInit(V3D *v3d)
{
  v3d->v[0] = v3d->v[1] = v3d->v[2] = 0.;
  v3d->v[3] = 1.;
  return v3d;
}


/*
 * v3dDelete ------------------------------------------------------------
 *
 * Free an malloc'ed T3D.
 */

void
v3dDelete(V3D *v3d)
{
   free(v3d);
}


/* --- v3dNorm ------------------------------------------------------- */

/** \brief Compute euclidian norm of a V3D
 *
 * \param[in] v3d	Input vector
 *
 * \return the euclidian norm of the vector.
 */

double
v3dNorm(const V3D *v3d)
{
  double d = sqrt(v3d->v[0]*v3d->v[0] +
		  v3d->v[1]*v3d->v[1] +
		  v3d->v[2]*v3d->v[2]);

  return (v3d->v[3] == 1.) ? d : d/v3d->v[3];
}


/* --- v3dNormalize -------------------------------------------------- */

/** \brief Normalize a V3D
 *
 * \param[inout] v3d	Input vector.
 *
 * \return the normalized vector.
 */

void
v3dNormalize(V3D *v3d)
{
  double d = v3dNorm(v3d);

  v3d->v[0] /= d;
  v3d->v[1] /= d;
  v3d->v[2] /= d;
}


/* --- v3dApplyT3D --------------------------------------------------- */

/** \brief Apply T3D to a V3D
 *
 * \param[in] t3d	Transformation.
 * \param[in] v3d	Vector to be transformed.
 *
 * \return the transformed vector (equal to `v3d').
 */

V3D *
v3dApplyT3D(T3D *t3d, V3D *v3d)
{
  V3D tmp;
  double *c, *v;
  int i, j;

  if (!t3dConvertTo(T3D_MATRIX, t3d)) return NULL;

  c = t3d->matrix.matrix;
  for(i=0; i<4; i++) {
    tmp.v[i] = 0.;
    v = v3d->v;
    for(j=0; j<4; j++) tmp.v[i] += *c++ * *v++;
  }

  *v3d = tmp;
  return v3d;
}


/* -- v3dProd -------------------------------------------------------- */

/** \brief Vector product
 *
 * \param[in] a		First vector
 * \param[in] b		Second vector
 *
 * \return a pointer on `a', containing the result of the
 * dot product.
 */

double 
v3dDot(const V3D *a, const V3D *b)
{
  double r;

  if (!v3dDotIn(&r, a, b)) return 0;
  return r;
}


/* -- v3dProdIn ------------------------------------------------------ */

/** \brief Vector product
 *
 * \param[out] c	Resulting vector
 * \param[in] a		First vector
 * \param[in] b		Second vector
 *
 * \return c, containing the result of the cross product.
 */

double *
v3dDotIn(double *c, const V3D *a, const V3D *b)
{
  double d;
  d =  a->v[0] * b->v[0];
  d += a->v[1] * b->v[1];
  d += a->v[2] * b->v[2];
  *c = d;
  return c;
}


/* -- v3dProd -------------------------------------------------------- */

/** \brief Vector product
 *
 * \param[in] a		First vector
 * \param[in] b		Second vector
 *
 * \return a pointer on `a', containing the result of the
 * cross product.
 */

V3D *
v3dProd(V3D *a, const V3D *b)
{
  V3D r;

  if (!v3dProdIn(&r, a, b)) return NULL;
  *a = r;
  return a;
}


/* -- v3dProdIn ------------------------------------------------------ */

/** \brief Vector product
 *
 * \param[out] c	Resulting vector
 * \param[in] a		First vector
 * \param[in] b		Second vector
 *
 * \return c, containing the result of the cross product.
 */

V3D *
v3dProdIn(V3D *c, const V3D *a, const V3D *b)
{
  c->v[0] = a->v[1]*b->v[2] - a->v[2]*b->v[1];
  c->v[1] = a->v[2]*b->v[0] - a->v[0]*b->v[2];
  c->v[2] = a->v[0]*b->v[1] - a->v[1]*b->v[0];
  c->v[3] = 1;
  return c;
}
