/*	$LAAS: t3dvector.c,v 1.1 2005/08/31 21:56:52 sara Exp $ */

/*
 * Copyright (c) 2000-2004 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <math.h>

#include "t3d/t3d.h"
#include "t3d_internal.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif /*M_PI*/

/* --- t3dVectorOfMatrix --------------------------------------------- */

/* Compute rotation vector from matrix */

void
t3dVectorOfMatrix(double *m, double *vector)
{
   double rx, ry, rz;
   double ryz, rzx, rxy;
   double co, t, f, g, max;

#define EPS	(.1)

   co = (m[0] + m[5] + m[10] - 1.)/2.0;
   if (co > 1.0) co = 1.0; else if (co < -1.0) co = -1.0;
   t = acos(co);

   if ((1+co) > EPS ) {
      /* theta is not close to PI */

      if (t > 1e-10) {
	 f = sin(t)/t;
	 rx = (m[9] - m[6])/2.0/f;
	 ry = (m[2] - m[8])/2.0/f;
	 rz = (m[4] - m[1])/2.0/f;
      } else {
	 /* teta = 0, identity transform */
	 f = 1.0;
	 rx = ry = rz = 0.0;
      }
   } else {
      /* theta is close to PI */

      if (1+co < 1e-10)  {  /* theta == PI */
	 g = (1-co)/t/t; /* g = 2/PI/PI */
	 rx = sqrt((m[0] - co)/g);
	 ry = sqrt((m[5] - co)/g);
	 rz = sqrt((m[10] - co)/g);

	 ryz = (m[9] + m[6])/2.0/g;
	 rzx = (m[2] + m[8])/2.0/g;
	 rxy = (m[4] + m[1])/2.0/g;

	 if (rx > ry) max = rx; else max = ry;
	 if (max < rz) max = rz;

	 if (rx==max) {
	    if ( rxy < 0.0 ) ry = -ry;
	    if ( rzx < 0.0 ) rz = -rz;
	 } else if (ry==max) {
	    if ( rxy < 0.0 ) rx = -rx;
	    if ( ryz < 0.0 ) rz = -rz;
	 } else {
	    if ( rzx < 0.0 ) rx = -rx;
	    if ( ryz < 0.0 ) ry = -ry;
	 }

      } else {	/* theta close to PI, but not PI */
	 rx = (m[9] - m[6])/2.0;
	 ry = (m[2] - m[8])/2.0;
	 rz = (m[4] - m[1])/2.0;

	 if (fabs(rx) > fabs(ry)) max = fabs(rx); else max = fabs(ry);
	 if (max < fabs(rz)) max = fabs(rz);

	 if (fabs(rx) == max) {
	    if ( rx < 0.0 ) t = 2*M_PI - t;
	 } else if (fabs(ry) == max) {
	    if ( ry < 0.0 ) t = 2*M_PI - t;
	 } else if (fabs(rz) == max) {
	    if ( rz < 0.0 ) t = 2*M_PI - t;
	 }

	 f = sin(t)/t;
	 rx /= f;
	 ry /= f;
	 rz /= f;
      }
   }

   t3d_dm_fill(vector, 1, 6, rx, ry, rz, m[3], m[7], m[11]);
}


/*
 * --- t3dDerivativeOfVector --------------------------------------------
 *
 * Compute derivative rotation matrices of vector angles
 */

void
t3dDerivativeOfVector(double *vector, double *dRx, double *dRy, double *dRz)
{
   double rx = vector[0];
   double ry = vector[1];
   double rz = vector[2];

   double t, t2;

   t = sqrt(t2 = (rx*rx + ry*ry + rz*rz));
   /* if teta = 0, vector is identity */
   if ( t > 1e-10 ) {
      double f = sin(t)/t;
      double c = cos(t);
      double g = (1-c)/t2;

      double fo = (c - f)/t2;
      double go = (f - 2*g)/t2;
      double co = -f;

      t3d_dm_fill(
	 dRx, 3, 3,
	 rx*co + rx*rx*rx*go + 2*rx*g,
	 rx*ry*rx*go + ry*g - rz*rx*fo,
	 rx*rz*rx*go + rz*g + ry*rx*fo,

	 rx*ry*rx*go + ry*g + rz*rx*fo,
	 rx*co + ry*ry*rx*go,
	 ry*rz*rx*go - rx*rx*fo - f,

	 rx*rz*rx*go + rz*g - ry*rx*fo,
	 ry*rz*rx*go + rx*rx*fo + f,
	 rx*co + rz*rz * rx*go);

      t3d_dm_fill(
	 dRy, 3, 3,
	 ry*co + rx*rx*ry*go,
	 rx*ry*ry*go + rx*g - rz*ry*fo,
	 rx*rz*ry*go + ry*ry*fo + f,

	 rx*ry*ry*go + rx*g + rz*ry*fo,
	 ry*co + ry*ry*ry*go + 2*ry*g,
	 ry*rz*ry*go + rz*g - rx*ry*fo,

	 rx*rz*ry*go - ry*ry*fo - f,
	 ry*rz*ry*go + rz*g + rx*ry*fo,
	 ry*co + rz*rz*ry*go );

      t3d_dm_fill(
	 dRz, 3, 3,
	 rz*co + rx*rx*rz*go,
	 rx*ry*rz*go - rz*rz*fo - f,
	 rx*rz*rz*go + rx*g + ry*rz*fo,

	 rx*ry*rz*go + rz*rz*fo + f,
	 rz*co + ry*ry*rz*go,
	 ry*rz*rz*go + ry*g - rx*rz*fo,

	 rx*rz*rz*go + rx*g - ry*rz*fo,
	 ry*rz*rz*go + ry*g + rx*rz*fo,
	 rz*co + rz*rz*rz*go + 2*rz*g);

   } else {
      /* f -> 1, g -> 1/2 , c -> 1, fo -> -1/3, go -> 1/4, co -> -1 */

      t3d_dm_fill(
	 dRx, 3, 3,
	 0.0, 0.0, 0.0,
	 0.0, 0.0, -1.0,
	 0.0, 1.0, 0.0);

      t3d_dm_fill(
	 dRy, 3, 3,
	 0.0, 0.0, 1.0,
	 0.0, 0.0, 0.0,
	 -1.0, 0.0, 0.0);

      t3d_dm_fill(
	 dRz, 3, 3,
	 0.0, -1.0, 0.0,
	 1.0, 0.0, 0.0,
	 0.0, 0.0, 0.0);
   }
}
