/* -*- c++ -*-
 *
 * Copyright (c) 2000-2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#ifndef H_T3D
#define H_T3D
#ifdef __cplusplus
# include <iostream>
# include <cstdio>
#else
# include <stdio.h>
#endif

/* Forward declarations */
typedef struct V3D		V3D;
typedef struct T3DMatrix	T3DMatrix;
typedef struct T3DUMatrix	T3DUMatrix;
typedef struct T3DBryan		T3DBryan;
typedef struct T3DUBryan	T3DUBryan;
typedef struct T3DVector	T3DVector;
typedef struct T3DUVector	T3DUVector;

typedef union  R3D		R3D;
typedef union  T3D		T3D;


/* --- C API --------------------------------------------------------- */

/** \addtogroup C */
/** \{ */

/* T3D error codes */
typedef enum T3DError {
#define T3D_MAGIC 0x3580
   T3D_EOK =			T3D_MAGIC | 0,
   T3D_EBADTYPE =		T3D_MAGIC | 1,
   T3D_ETYPESDIFFER =		T3D_MAGIC | 2,
   T3D_EBADFLAGS =		T3D_MAGIC | 3,
   T3D_ESIZE =			T3D_MAGIC | 4,
   T3D_ENOTIMPLEMENTED =	T3D_MAGIC | 15
} T3DError;

/** \brief Internal representation of a T3D.
 *
 * This value can be supplied to t3dNew() or T3D() contstructor.
 */
typedef enum T3DType {
  T3D_MATRIX,		/**<\brief Homogeneous matrix */
  T3D_U_MATRIX,		/**<\brief Uncertain homogeneous matrix. */
  T3D_BRYAN,		/**<\brief Bryan angles and translation. */
  T3D_U_BRYAN,		/**<\brief Uncertain bryan angles and translation. */
  T3D_VECTOR,		/**<\brief Rotation vector */
  T3D_U_VECTOR		/**<\brief Uncertain rotation vector */
} T3DType;

/** \brief Tait-Bryan angle names
 *
 * Use this enumerator with Tait-Bryan angle accessors.
 */
typedef enum T3DBryanIndex {
  T3D_BRYAN_YAW	=	0,	/**<\brief Yaw angle or Z-axis rotation. */
  T3D_BRYAN_PITCH =	1,	/**<\brief Pitch angle or Y-axis rotation. */
  T3D_BRYAN_ROLL =	2,	/**<\brief Roll angle or X-axis rotation. */
  T3D_BRYAN_X	=	3,	/**<\brief Translation on X-axis. */
  T3D_BRYAN_Y =		4,	/**<\brief Translation on Y-axis. */
  T3D_BRYAN_Z =		5	/**<\brief Translation on Z-axis. */
} T3DBryanIndex;

/* T3D flags */
typedef enum T3DFlags {
   T3D_MINIMAL_SIZE =		1<<0,
   T3D_ALLOW_CONVERSION =	1<<1
} T3DFlags;

extern const T3D t3dBryanIdentity;	/**<\brief static initializer. */
extern const T3D t3dUBryanIdentity;	/**<\brief static initializer. */
extern const T3D t3dVectorIdentity;	/**<\brief static initializer. */
extern const T3D t3dUVectorIdentity;	/**<\brief static initializer. */
extern const T3D t3dMatrixIdentity;	/**<\brief static initializer. */
extern const T3D t3dUMatrixIdentity;	/**<\brief static initializer. */

#ifdef __cplusplus
extern "C" {
#endif

  T3D *		t3dNew(T3DType type, int flags);
  T3D *		t3dInit(T3D *t3d, T3DType type, int flags);
  T3D *		t3dDup(T3D *t3d);
  void		t3dDelete(T3D *t3d);

  T3D *		t3dCopyIn(T3D *cpy, const T3D *src);
  T3D *		t3dInvert(T3D *t3d);
  T3D *		t3dInvertIn(T3D *inv, T3D *t3d);
  T3D *		t3dConvertTo(T3DType typeTo, T3D *t3d);

  T3D *		t3dComp(T3D *a, T3D *b);
  T3D *		t3dCompIn(T3D *c, T3D *a, T3D *b);

  int		t3dFPrint(FILE *stream, const T3D *t3d);

  static inline void	t3dWarn(const T3D *t3d);
  static inline int	t3dUncertain(T3DType type);
  static inline int	t3dOk(const T3D *t3d);
  static inline void	t3dSetOk(T3D *t3d);
  static inline T3D *	t3dSetError(T3D *t3d, int err);

#ifdef __cplusplus
}
#endif

/** \} */

/* --- Internal representations definition --------------------------- */

/** common fields for all representations */
#define _T3DHeader_members					\
  T3DType type;	/**<\brief internal representation type. */	\
  short flags;	/**<\brief special behavior. */			\
  int err	/**\brief last error code. */

/* Tait-Bryan angles */
#define _T3DBryan_members						\
  union /* anonymous */ {						\
    double bryan[6];	/**<\brief yaw, pitch, roll, x, y, z, in that order. */ \
    									\
    struct /* anonymous */ {						\
      double yaw;	/**<\brief yaw angle, rotation around Z. */	\
      double pitch;	/**<\brief pitch angle, rotation around Y. */	\
      double roll;	/**<\brief roll angle, rotation around X. */	\
      double x;		/**<\brief translation along X. */		\
      double y;		/**<\brief translation along Y. */		\
      double z;		/**<\brief translation along Z. */		\
    };									\
									\
    double euler[6];	/**<\deprecated see #T3DBryan::bryan. */	\
  }

/* Rotation vector */
#define _T3DVector_members						\
  union /* anonymous */ {						\
    double vector[6];	/**<\brief rx, ry, rz, x, y, z. */		\
									\
    struct /* anonymous */ {						\
      double r[3];	/**<\brief rotation vector. */			\
      double t[3];	/**<\brief translation vector. */		\
    };									\
									\
    struct /* anonymous */ {						\
      double rx;	/**<\brief rotation around X. */		\
      double ry;	/**<\brief rotation around Y. */		\
      double rz;	/**<\brief rotation around Z. */		\
      double x;		/**<\brief translation along X. */		\
      double y;		/**<\brief translation along Y. */		\
      double z;		/**<\brief translation along Z. */		\
    };									\
  }


/* --- T3D data structures ------------------------------------------- */


/** \brief Homogeneous matrix.
 *
 * Internal representation of a T3D of type ::T3D_MATRIX.
 */
struct T3DMatrix {
#ifdef __cplusplus
  inline double& operator[](unsigned int i);
  inline double operator[](unsigned int i) const;
  inline double& operator()(unsigned int i, unsigned int j);
  inline double operator()(unsigned int i, unsigned int j) const;
#endif

  _T3DHeader_members;
  double matrix[16];		/**<\brief Homogeneous matrix. */
};

/** \brief Uncertain Homogeneous matrix.
 *
 * Internal representation of a T3D of type ::T3D_U_MATRIX.
 */
struct T3DUMatrix {
#ifdef __cplusplus
  inline double operator[](int i);
#endif

  _T3DHeader_members;
  double matrix[16];		/**<\brief Homogeneous matrix. */
  double v[21];			/**<\brief Covariance matrix on _Bryan_ angles. */
};

/** \brief Tait-Bryan angles and translation.
 *
 * Internal representation of a T3D of tupe ::T3D_BRYAN.
 */
struct T3DBryan {
#ifdef __cplusplus
  inline double &operator[](int i);
  inline double &operator[](T3DBryanIndex i);
#endif

  _T3DHeader_members;
  _T3DBryan_members;
};

/** \brief Uncertain Tait-Bryan angles and translation.
 *
 * Internal representation of a T3D of tupe ::T3D_U_BRYAN.
 */
struct T3DUBryan {
#ifdef __cplusplus
  inline double &operator[](int i);
  inline double &operator[](T3DBryanIndex i);
#endif

  _T3DHeader_members;
  _T3DBryan_members;
  double v[21];		/**<\brief covariance matrix on angles and
			 * translation. */
};

/** \brief Rotation vector and translation.
 *
 * Internal representation of a T3D of tupe ::T3D_VECTOR.
 */
struct T3DVector {
  _T3DHeader_members;
  _T3DVector_members;
};

/** \brief Uncertain rotation vector and translation.
 *
 * Internal representation of a T3D of tupe ::T3D_U_VECTOR.
 */
struct T3DUVector {
  _T3DHeader_members;
  _T3DVector_members;
  double v[21];		/**<\brief covariance matrix on the above */
};


/** \brief Abstract representation.
 *
 * Internal representation with abstract meaning. It is guaranteed to be
 * able to contain any other representation.
 */
typedef struct T3DAny {
  _T3DHeader_members;
  double pad[37];
} T3DAny;


/** \brief A 3D affine transformation
 *
 * This union is the base class of affine transforms. It can hold
 * any representation, but only one at a time. Implicits or explicits
 * conversions can be performed.
 */
union T3D {
#ifdef __cplusplus
   inline T3D(T3DType type, int flags, T3DError error,const double* d1 =0x0,const double* d2 = 0x0);
   inline T3D(T3DType type = T3D_MATRIX, int flags = 0);
   inline T3D(const T3D *t3d);
   inline T3D(const T3DMatrix &m);
   inline T3D(double _yaw, double _pitch, double _roll,
	      double _x, double _y, double _z);

   inline T3D &invert(void);
   inline T3D &convertTo(T3DType type);

   inline T3D operator~(void) const;
   inline T3D operator*(T3D &t3d);

   friend inline V3D operator*(T3D &t3d, const V3D &b);
   friend std::ostream& operator<<(std::ostream &os, const T3D &t3d);

#endif

  struct /* anonymous */ { _T3DHeader_members; };

  T3DMatrix	matrix;		/**<\brief Homogeneous matrix. */
  T3DUMatrix	umatrix;	/**<\brief Uncertain homogeneous matrix. */
  T3DBryan	bryan;		/**<\brief Tait-Bryan angles and translation. */
  T3DUBryan	ubryan;		/**<\brief Uncertain Tait-Bryan. */
  T3DVector	vector;		/**<\brief Rotation vector. */
  T3DUVector	uvector;	/**<\brief Uncertain rotation vector. */

  T3DUBryan	euler;		/**<\deprecated use #T3D::ubryan. */
};

/* include inlined functions and methods */
# include "t3d/t3d_inline.h"

/* backward compatibility */
#include "t3d/deprecated.h"

#endif /* H_T3D */
