/* -*- c++ -*-
 *
 * Copyright (c) 2007 Anthony Mallet                  --  Sat Feb 17 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#ifndef H_T3D_INLINE
#define H_T3D_INLINE

/* Inlined functions definition */

/** \addtogroup C */
/** \{ */

/* --- t3dWarn ------------------------------------------------------- */

/** \brief Echo a message if a t3d has its errno set.
 *
 * The function is made public so that you can easily insert a breakpoint
 * from within a debugger.
 * \param[in] t3d	T3D object.
 */
static inline void
t3dWarn(const T3D *t3d)
{
  fputs("t3d: error while manipulating a T3D\n", stderr);
}


/* --- t3dUncertain -------------------------------------------------- */

/** \brief Check if a T3D type is uncertain
 *
 * \param[in] type	type to check.
 * \return true if type is uncertain, false otherwise.
 */
static inline int
t3dUncertain(T3DType type)
{
  return
    type == T3D_U_MATRIX ||
    type == T3D_U_BRYAN ||
    type == T3D_U_VECTOR;
}


/* --- t3dOk --------------------------------------------------------- */

/** \brief Check if t3d has an error.
 *
 * \param[in] t3d	T3D object.
 * \return true if t3d has no error, false otherwise.
 */
static inline int
t3dOk(const T3D *t3d)
{
  return t3d && t3d->err == T3D_EOK;
}


/* --- t3dSetOk ------------------------------------------------------ */

/** \brief Reset t3d error code.
 *
 * \param[in] t3d	T3D object.
 */
static inline void
t3dSetOk(T3D *t3d)
{
  t3d->err = T3D_EOK;
}


static inline T3D *
t3dSetError(T3D *t3d, int err)
{
  t3d->err = err;
  t3dWarn(t3d);
  return NULL;
}

/** \} */


#ifdef __cplusplus

/** \addtogroup CXX */
/** \{ */

/* --- T3DMatrix::operator[] ----------------------------------------- */

/** \brief Acessor for matrices elements.
 *
 * The function return the element i of the homogeneous matrix. Elements
 * are numbered in 'row-major' mode.
 *
 * \param[in] i		Matrix element number.
 * \return reference to element.
 */

inline double &
T3DMatrix::operator[](unsigned int i)
{
  return matrix[i];
}

inline double
T3DMatrix::operator[](unsigned int i) const
{
  return matrix[i];
}


/* --- T3DMatrix::operator(,) ----------------------------------------- */

/** \brief Acessor for matrices elements.
 *
 * The function return the element (i,j) of the homogeneous matrix. 
 *
 * \param[in] i,j		Matrix element indices.
 * \return reference to element.
 */

inline double &
T3DMatrix::operator()(unsigned int i, unsigned int j)
{
  return matrix[4*i+j];
}

inline double
T3DMatrix::operator()(unsigned int i, unsigned int j) const
{
  return matrix[4*i+j];
}


/* --- T3DBryan::operator[] ------------------------------------------ */

/** \brief Accessor for Tait-Brayan angles.
 *
 * \param[in] i		Angle number, \sa #T3DBryanIndex.
 * \return a reference on the angle value.
 */

inline double &
T3DBryan::operator[](int i)
{
  return euler[i];
}

/** \copydoc T3DBryan::operator[](int i) */
inline double &
T3DBryan::operator[](T3DBryanIndex i)
{
  return euler[i];
}



/* --- T3D::T3D ------------------------------------------------------ */

/** \brief T3D constructor
 *
 * Initialize a 3D transformation to Identity.
 *
 * \param[in] type	Internal representation.
 * \param[in] flags	Optional parameters to alter the default behavior
 */
inline 
T3D::T3D(T3DType _type, int _flags, T3DError error,const double* d1,const double* d2)
{
  type = _type;
  flags = _flags;
  err = error;
	switch (type)
	{
	case(T3D_MATRIX):
		{
			for (unsigned i = 0; i < 16; ++i) matrix.matrix[i] = d1[i];
			break;
		}
	case(T3D_U_MATRIX):
		{
			for (unsigned i = 0; i < 16; ++i) umatrix.matrix[i] = d1[i];
			for (unsigned i = 0; i < 21; ++i) umatrix.v[i]  = d2[i];
			break;
		}
	case(T3D_BRYAN):
		{
			for (unsigned i = 0; i < 6; ++i) bryan.bryan[i] = d1[i];
			break;
		}
	case(T3D_U_BRYAN):
		{
			for (unsigned i = 0; i < 6; ++i) ubryan.bryan[i] = d1[i];
			for (unsigned i = 0; i < 21; ++i) ubryan.v[i] = d2[i];
			break;
		}
	case(T3D_VECTOR):
		{
			for (unsigned i = 0; i < 6; ++i) vector.vector[i] = d1[i];
			break;
		}
	case(T3D_U_VECTOR):
		{
			for (unsigned i = 0; i < 6; ++i) uvector.vector[i] = d1[i];
			for (unsigned i = 0; i < 21; ++i) uvector.v[i] = d2[i];
			break;
		}
	}
}

/* --- T3D::T3D ------------------------------------------------------ */

/** \brief T3D constructor
 *
 * Initialize a 3D transformation to Identity.
 *
 * \param[in] type	Internal representation.
 * \param[in] flags	Optional parameters to alter the default behavior
 */

inline
T3D::T3D(T3DType type, int flags)
{
  t3dInit(this, type, flags);
}


/* --- T3D::T3D ------------------------------------------------------ */

/** \brief T3D constructor
 *
 * Initialize a 3D transformation equal to t3d.
 *
 * \param[in] t3d	Copied transformation.
 */

inline
T3D::T3D(const T3D *t3d)
{
  t3dCopyIn(this, t3d);
}

/** \copydoc T3D::T3D(const T3D *) */
inline
T3D::T3D(const T3DMatrix &t3d)
{
  matrix = t3d;
}


/* --- T3DBryan::T3DBryan -------------------------------------------- */

/** \brief T3DBryan constructor
 *
 * Initialize a 3D transformation with the given angles and translations.
 *
 * \param[in] yaw	Rotation around Z.
 * \param[in] pitch	Rotation around Y.
 * \param[in] roll	Rotation around X.
 * \param[in] x		Translation along X.
 * \param[in] y		Translation along Y.
 * \param[in] z		Translation along Z.
 */

inline
T3D::T3D(double yaw, double pitch, double roll,
	 double x, double y, double z)
{
  t3dInit(this, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  bryan.yaw = yaw;
  bryan.pitch = pitch;
  bryan.roll = roll;
  bryan.x = x;
  bryan.y = y;
  bryan.z = z;
}


/* --- T3D::convertTo ------------------------------------------------ */

/** \brief Convert a T3D into another representation.
 *
 * \sa t3dConvertTo().
 *
 * \param[in] type	Type to convert to.
 */

inline T3D &
T3D::convertTo(T3DType type)
{
  return *t3dConvertTo(type, this);
}


/* --- T3D::invert --------------------------------------------------- */

/** \brief Invert a T3D.
 *
 * Invert t3d object. \sa t3dInvert().
 */

inline T3D &
T3D::invert()
{
  return *t3dInvert(this);
}


/* --- T3D::operator~ ------------------------------------------------ */

/** \brief Inverse of a T3D.
 *
 * Return the inverse of t3d. \sa t3dInvert().
 */

inline T3D
T3D::operator~() const
{
  T3D r(this);
  return *t3dInvert(&r);
}


/* --- T3D::operator* ------------------------------------------------ */

/** \brief Composition of two T3D.
 *
 * In the expression this*t3d, return the composition of t3d by this. \sa
 * t3dComp().
 *
 * \param[in] t3d	First transformation.
 */

inline T3D
T3D::operator*(T3D &t3d)
{
  T3D r(t3d);
  return *t3dComp(&r, this);
}

/** \} */
#endif /* __cplusplus */

#endif /* H_T3D_INLINE */
